@@ -23,11 +23,12 @@
 #include "tracker-direct.h"
 #include "tracker-direct-batch.h"
 #include "tracker-direct-statement.h"
-#include "libtracker-sparql/tracker-private.h"
-#include <libtracker-data/tracker-data.h>
-#include <libtracker-data/tracker-sparql.h>
+
+#include <libtracker-common/tracker-utils.h>
+#include <libtracker-sparql/core/tracker-data.h>
 #include <libtracker-sparql/tracker-notifier-private.h>
 #include <libtracker-sparql/tracker-private.h>
+#include <libtracker-sparql/tracker-serializer.h>
 
 typedef struct _TrackerDirectConnectionPrivate TrackerDirectConnectionPrivate;
 
@@ -60,6 +61,17 @@ typedef struct {
 	TrackerResource *resource;
 } UpdateResource;
 
+typedef struct {
+	gchar *query;
+	TrackerRdfFormat format;
+} SerializeRdf;
+
+typedef struct {
+	GInputStream *stream;
+	gchar *default_graph;
+	TrackerRdfFormat format;
+} DeserializeRdf;
+
 enum {
 	PROP_0,
 	PROP_FLAGS,
@@ -77,6 +89,8 @@ typedef enum {
 	TASK_TYPE_UPDATE_RESOURCE,
 	TASK_TYPE_UPDATE_BATCH,
 	TASK_TYPE_RELEASE_MEMORY,
+	TASK_TYPE_SERIALIZE,
+	TASK_TYPE_DESERIALIZE,
 } TaskType;
 
 typedef struct {
@@ -185,6 +199,19 @@ error:
 	return FALSE;
 }
 
+static TrackerSerializerFormat
+convert_format (TrackerRdfFormat format)
+{
+	switch (format) {
+	case TRACKER_RDF_FORMAT_TURTLE:
+		return TRACKER_SERIALIZER_FORMAT_TTL;
+	case TRACKER_RDF_FORMAT_TRIG:
+		return TRACKER_SERIALIZER_FORMAT_TRIG;
+	default:
+		g_assert_not_reached ();
+	}
+}
+
 static void
 update_thread_func (gpointer data,
                     gpointer user_data)
@@ -207,6 +234,7 @@ update_thread_func (gpointer data,
 
 	switch (task_data->type) {
 	case TASK_TYPE_QUERY:
+	case TASK_TYPE_SERIALIZE:
 		g_warning ("Queries don't go through this thread");
 		break;
 	case TASK_TYPE_UPDATE:
@@ -221,6 +249,30 @@ update_thread_func (gpointer data,
 		update_resource (tracker_data, data->graph, data->resource, &error);
 		break;
 	}
+	case TASK_TYPE_DESERIALIZE: {
+		DeserializeRdf *data = task_data->data;
+		TrackerSparqlCursor *deserializer;
+
+		tracker_data_begin_transaction (tracker_data, &error);
+		if (error)
+			break;
+
+		deserializer = tracker_deserializer_new (data->stream,
+		                                         priv->namespace_manager,
+		                                         convert_format (data->format));
+
+		if (tracker_data_load_from_deserializer (tracker_data,
+		                                         TRACKER_DESERIALIZER (deserializer),
+		                                         data->default_graph,
+		                                         "<stream>",
+		                                         &error)) {
+			tracker_data_commit_transaction (tracker_data, &error);
+		} else {
+			tracker_data_rollback_transaction (tracker_data);
+		}
+		g_object_unref (deserializer);
+		break;
+	}
 	case TASK_TYPE_UPDATE_BATCH:
 		tracker_direct_batch_update (task_data->data, priv->data_manager, &error);
 		break;
@@ -245,18 +297,78 @@ update_thread_func (gpointer data,
 	g_mutex_unlock (&priv->mutex);
 }
 
+static void
+execute_query_in_thread (GTask    *task,
+                         TaskData *task_data)
+{
+	TrackerSparqlCursor *cursor;
+	GError *error = NULL;
+
+	cursor = tracker_sparql_connection_query (TRACKER_SPARQL_CONNECTION (g_task_get_source_object (task)),
+	                                          task_data->data,
+	                                          g_task_get_cancellable (task),
+	                                          &error);
+	if (cursor)
+		g_task_return_pointer (task, cursor, g_object_unref);
+	else
+		g_task_return_error (task, error);
+}
+
+static void
+serialize_in_thread (GTask    *task,
+                     TaskData *task_data)
+{
+	TrackerDirectConnectionPrivate *priv;
+	TrackerDirectConnection *conn;
+	TrackerSparql *query = NULL;
+	TrackerSparqlCursor *cursor = NULL;
+	TrackerNamespaceManager *namespaces;
+	GInputStream *istream = NULL;
+	SerializeRdf *data = task_data->data;
+	GError *error = NULL;
+
+	conn = g_task_get_source_object (task);
+	priv = tracker_direct_connection_get_instance_private (conn);
+
+	g_mutex_lock (&priv->mutex);
+	query = tracker_sparql_new (priv->data_manager, data->query);
+	if (!tracker_sparql_is_serializable (query)) {
+		g_set_error (&error,
+		             TRACKER_SPARQL_ERROR,
+		             TRACKER_SPARQL_ERROR_PARSE,
+		             "Query is not DESCRIBE or CONSTRUCT");
+		goto out;
+	}
+
+	cursor = tracker_sparql_execute_cursor (query, NULL, &error);
+	tracker_direct_connection_update_timestamp (conn);
+	if (!cursor)
+		goto out;
+
+	tracker_sparql_cursor_set_connection (cursor, TRACKER_SPARQL_CONNECTION (conn));
+	namespaces = tracker_sparql_connection_get_namespace_manager (TRACKER_SPARQL_CONNECTION (conn));
+	istream = tracker_serializer_new (cursor, namespaces,
+	                                  convert_format (data->format));
+
+ out:
+	g_clear_object (&query);
+	g_clear_object (&cursor);
+	g_mutex_unlock (&priv->mutex);
+
+	if (istream)
+		g_task_return_pointer (task, istream, g_object_unref);
+	else
+		g_task_return_error (task, error);
+}
+
 static void
 query_thread_pool_func (gpointer data,
                         gpointer user_data)
 {
 	TrackerDirectConnection *conn = user_data;
 	TrackerDirectConnectionPrivate *priv;
-	TrackerSparqlCursor *cursor;
 	GTask *task = data;
 	TaskData *task_data = g_task_get_task_data (task);
-	GError *error = NULL;
-
-	g_assert (task_data->type == TASK_TYPE_QUERY);
 
 	priv = tracker_direct_connection_get_instance_private (conn);
 
@@ -269,14 +381,16 @@ query_thread_pool_func (gpointer data,
 		return;
 	}
 
-	cursor = tracker_sparql_connection_query (TRACKER_SPARQL_CONNECTION (g_task_get_source_object (task)),
-	                                          task_data->data,
-	                                          g_task_get_cancellable (task),
-	                                          &error);
-	if (cursor)
-		g_task_return_pointer (task, cursor, g_object_unref);
-	else
-		g_task_return_error (task, error);
+	switch (task_data->type) {
+	case TASK_TYPE_QUERY:
+		execute_query_in_thread (task, task_data);
+		break;
+	case TASK_TYPE_SERIALIZE:
+		serialize_in_thread (task, task_data);
+		break;
+	default:
+		g_assert_not_reached ();
+	}
 
 	g_object_unref (task);
 }
@@ -331,6 +445,8 @@ translate_flags (TrackerSparqlConnectionFlags flags)
 		db_flags |= TRACKER_DB_MANAGER_FTS_ENABLE_STOP_WORDS;
 	if ((flags & TRACKER_SPARQL_CONNECTION_FLAGS_FTS_IGNORE_NUMBERS) != 0)
 		db_flags |= TRACKER_DB_MANAGER_FTS_IGNORE_NUMBERS;
+	if ((flags & TRACKER_SPARQL_CONNECTION_FLAGS_ANONYMOUS_BNODES) != 0)
+		db_flags |= TRACKER_DB_MANAGER_ANONYMOUS_BNODES;
 
 	return db_flags;
 }
@@ -452,7 +568,9 @@ get_event_cache_ht (TrackerNotifier *notifier)
 
 	events = g_object_get_qdata (G_OBJECT (notifier), tracker_direct_notifier_quark ());
 	if (!events) {
-		events = g_hash_table_new_full (NULL, NULL, NULL,
+		events = g_hash_table_new_full (tracker_rowid_hash,
+		                                tracker_rowid_equal,
+		                                (GDestroyNotify) tracker_rowid_free,
 		                                (GDestroyNotify) _tracker_notifier_event_cache_free);
 		g_object_set_qdata_full (G_OBJECT (notifier), tracker_direct_notifier_quark (),
 		                         events, (GDestroyNotify) g_hash_table_unref);
@@ -463,18 +581,18 @@ get_event_cache_ht (TrackerNotifier *notifier)
 
 static TrackerNotifierEventCache *
 lookup_event_cache (TrackerNotifier *notifier,
-                    gint             graph_id,
+                    TrackerRowid     graph_id,
                     const gchar     *graph)
 {
 	TrackerNotifierEventCache *cache;
 	GHashTable *events;
 
 	events = get_event_cache_ht (notifier);
-	cache = g_hash_table_lookup (events, GINT_TO_POINTER (graph_id));
+	cache = g_hash_table_lookup (events, &graph_id);
 
 	if (!cache) {
 		cache = _tracker_notifier_event_cache_new (notifier, graph);
-		g_hash_table_insert (events, GINT_TO_POINTER (graph_id), cache);
+		g_hash_table_insert (events, tracker_rowid_copy (&graph_id), cache);
 	}
 
 	return cache;
@@ -484,13 +602,11 @@ lookup_event_cache (TrackerNotifier *notifier,
  * (always the same one though), handle with care.
  */
 static void
-insert_statement_cb (gint         graph_id,
-                     const gchar *graph,
-                     gint         subject_id,
-                     const gchar *subject,
-                     gint         predicate_id,
-                     gint         object_id,
-                     const gchar *object,
+insert_statement_cb (TrackerRowid  graph_id,
+                     const gchar  *graph,
+                     TrackerRowid  subject_id,
+                     TrackerRowid  predicate_id,
+                     TrackerRowid  object_id,
                      GPtrArray   *rdf_types,
                      gpointer     user_data)
 {
@@ -502,7 +618,7 @@ insert_statement_cb (gint         graph_id,
 	TrackerProperty *rdf_type = tracker_ontologies_get_rdf_type (ontologies);
 	TrackerNotifierEventCache *cache;
 	TrackerClass *new_class = NULL;
-	gint i;
+	guint i;
 
 	cache = lookup_event_cache (notifier, graph_id, graph);
 
@@ -530,13 +646,11 @@ insert_statement_cb (gint         graph_id,
 }
 
 static void
-delete_statement_cb (gint         graph_id,
-                     const gchar *graph,
-                     gint         subject_id,
-                     const gchar *subject,
-                     gint         predicate_id,
-                     gint         object_id,
-                     const gchar *object,
+delete_statement_cb (TrackerRowid  graph_id,
+                     const gchar  *graph,
+                     TrackerRowid  subject_id,
+                     TrackerRowid  predicate_id,
+                     TrackerRowid  object_id,
                      GPtrArray   *rdf_types,
                      gpointer     user_data)
 {
@@ -548,12 +662,15 @@ delete_statement_cb (gint         graph_id,
 	TrackerProperty *rdf_type = tracker_ontologies_get_rdf_type (ontologies);
 	TrackerNotifierEventCache *cache;
 	TrackerClass *class_being_removed = NULL;
-	gint i;
+	guint i;
 
 	cache = lookup_event_cache (notifier, graph_id, graph);
 
 	if (predicate_id == tracker_property_get_id (rdf_type)) {
-		class_being_removed = tracker_ontologies_get_class_by_uri (ontologies, object);
+		const gchar *uri;
+
+		uri = tracker_ontologies_get_uri_by_id (ontologies, object_id);
+		class_being_removed = tracker_ontologies_get_class_by_uri (ontologies, uri);
 	}
 
 	for (i = 0; i < rdf_types->len; i++) {
@@ -730,9 +847,6 @@ tracker_direct_connection_query (TrackerSparqlConnection  *self,
 	cursor = tracker_sparql_execute_cursor (query, NULL, &inner_error);
 	tracker_direct_connection_update_timestamp (conn);
 	g_object_unref (query);
-
-	if (cursor)
-		tracker_sparql_cursor_set_connection (cursor, self);
 	g_mutex_unlock (&priv->mutex);
 
 	if (inner_error)
@@ -846,6 +960,26 @@ tracker_direct_connection_update_finish (TrackerSparqlConnection  *self,
 		g_propagate_error (error, _translate_internal_error (inner_error));
 }
 
+static void
+on_batch_finished (GObject      *source,
+                   GAsyncResult *result,
+                   gpointer      user_data)
+{
+	TrackerBatch *batch = TRACKER_BATCH (source);
+	GTask *task = user_data;
+	GError *error = NULL;
+	gboolean retval;
+
+	retval = tracker_batch_execute_finish (batch, result, &error);
+
+	if (retval)
+		g_task_return_boolean (task, TRUE);
+	else
+		g_task_return_error (task, error);
+
+	g_object_unref (task);
+}
+
 static void
 tracker_direct_connection_update_array_async (TrackerSparqlConnection  *self,
                                               gchar                   **updates,
@@ -854,29 +988,18 @@ tracker_direct_connection_update_array_async (TrackerSparqlConnection  *self,
                                               GAsyncReadyCallback       callback,
                                               gpointer                  user_data)
 {
-	TrackerDirectConnectionPrivate *priv;
-	TrackerDirectConnection *conn;
-	TaskData *task_data;
+	TrackerBatch *batch;
 	GTask *task;
-	gchar *concatenated;
-	gchar **array_copy;
+	gint i;
 
-	conn = TRACKER_DIRECT_CONNECTION (self);
-	priv = tracker_direct_connection_get_instance_private (conn);
+	batch = tracker_sparql_connection_create_batch (self);
 
-	/* Make a NULL-terminated array and concatenate it */
-	array_copy = g_new0 (gchar *, n_updates + 1);
-	memcpy (array_copy, updates, n_updates * sizeof (gchar *));
-	concatenated = g_strjoinv ("\n", array_copy);
-	g_free (array_copy);
-
-	task_data = task_data_query_new (TASK_TYPE_UPDATE, concatenated, g_free);
+	for (i = 0; i < n_updates; i++)
+		tracker_batch_add_sparql (batch, updates[i]);
 
 	task = g_task_new (self, cancellable, callback, user_data);
-	g_task_set_task_data (task, task_data,
-	                      (GDestroyNotify) task_data_free);
-
-	g_thread_pool_push (priv->update_thread, task, NULL);
+	tracker_batch_execute_async (batch, cancellable, on_batch_finished, task);
+	g_object_unref (batch);
 }
 
 static gboolean
@@ -1176,6 +1299,184 @@ tracker_direct_connection_create_batch (TrackerSparqlConnection *connection)
 	return tracker_direct_batch_new (connection);
 }
 
+static gboolean
+tracker_direct_connection_lookup_dbus_service (TrackerSparqlConnection  *connection,
+                                               const gchar              *dbus_name,
+                                               const gchar              *dbus_path,
+                                               gchar                   **name,
+                                               gchar                   **path)
+{
+	TrackerDirectConnectionPrivate *priv;
+	TrackerDirectConnection *conn;
+	TrackerSparqlConnection *remote;
+	GError *error = NULL;
+	gchar *uri;
+
+	conn = TRACKER_DIRECT_CONNECTION (connection);
+	priv = tracker_direct_connection_get_instance_private (conn);
+
+	uri = tracker_util_build_dbus_uri (G_BUS_TYPE_SESSION,
+	                                   dbus_name, dbus_path);
+	remote = tracker_data_manager_get_remote_connection (priv->data_manager,
+	                                                     uri, &error);
+	if (error) {
+		g_warning ("Error getting remote connection '%s': %s", uri, error->message);
+		g_error_free (error);
+	}
+
+	g_free (uri);
+
+	if (!remote)
+		return FALSE;
+	if (!g_object_class_find_property (G_OBJECT_GET_CLASS (remote), "bus-name"))
+		return FALSE;
+
+	g_object_get (remote,
+	              "bus-name", name,
+	              "bus-object-path", path,
+	              NULL);
+
+	return TRUE;
+}
+
+static SerializeRdf *
+serialize_rdf_data_new (const gchar      *query,
+                        TrackerRdfFormat  format)
+{
+	SerializeRdf *data;
+
+	data = g_new0 (SerializeRdf, 1);
+	data->query = g_strdup (query);
+	data->format = format;
+
+	return data;
+}
+
+static void
+serialize_rdf_data_free (gpointer user_data)
+{
+	SerializeRdf *data = user_data;
+
+	g_free (data->query);
+	g_free (data);
+}
+
+static void
+tracker_direct_connection_serialize_async (TrackerSparqlConnection  *self,
+                                           TrackerSerializeFlags     flags,
+                                           TrackerRdfFormat          format,
+                                           const gchar              *query,
+                                           GCancellable             *cancellable,
+                                           GAsyncReadyCallback      callback,
+                                           gpointer                 user_data)
+{
+	TrackerDirectConnectionPrivate *priv;
+	TrackerDirectConnection *conn;
+	GError *error = NULL;
+	GTask *task;
+
+	conn = TRACKER_DIRECT_CONNECTION (self);
+	priv = tracker_direct_connection_get_instance_private (conn);
+
+	task = g_task_new (self, cancellable, callback, user_data);
+	g_task_set_task_data (task,
+	                      task_data_query_new (TASK_TYPE_SERIALIZE,
+	                                           serialize_rdf_data_new (query, format),
+	                                           serialize_rdf_data_free),
+	                      (GDestroyNotify) task_data_free);
+
+	if (!g_thread_pool_push (priv->select_pool, task, &error)) {
+		g_task_return_error (task, _translate_internal_error (error));
+		g_object_unref (task);
+	}
+}
+
+static GInputStream *
+tracker_direct_connection_serialize_finish (TrackerSparqlConnection  *connection,
+                                            GAsyncResult             *res,
+                                            GError                  **error)
+{
+	return g_task_propagate_pointer (G_TASK (res), error);
+}
+
+static DeserializeRdf *
+deserialize_rdf_data_new (GInputStream     *stream,
+                          const gchar      *default_graph,
+                          TrackerRdfFormat  format)
+{
+	DeserializeRdf *data;
+
+	data = g_new0 (DeserializeRdf, 1);
+	data->stream = g_object_ref (stream);
+	data->default_graph = g_strdup (default_graph);
+	data->format = format;
+
+	return data;
+}
+
+static void
+deserialize_rdf_data_free (gpointer user_data)
+{
+	DeserializeRdf *data = user_data;
+
+	g_object_unref (data->stream);
+	g_free (data->default_graph);
+	g_free (data);
+}
+
+static void
+tracker_direct_connection_deserialize_async (TrackerSparqlConnection *self,
+                                             TrackerDeserializeFlags  flags,
+                                             TrackerRdfFormat         format,
+                                             const gchar             *default_graph,
+                                             GInputStream            *stream,
+                                             GCancellable            *cancellable,
+                                             GAsyncReadyCallback      callback,
+                                             gpointer                 user_data)
+{
+	TrackerDirectConnectionPrivate *priv;
+	TrackerDirectConnection *conn;
+	GTask *task;
+
+	conn = TRACKER_DIRECT_CONNECTION (self);
+	priv = tracker_direct_connection_get_instance_private (conn);
+
+	task = g_task_new (self, cancellable, callback, user_data);
+	g_task_set_task_data (task,
+	                      task_data_query_new (TASK_TYPE_DESERIALIZE,
+	                                           deserialize_rdf_data_new (stream,
+	                                                                     default_graph,
+	                                                                     format),
+	                                           deserialize_rdf_data_free),
+	                      (GDestroyNotify) task_data_free);
+
+	g_thread_pool_push (priv->update_thread, task, NULL);
+}
+
+static gboolean
+tracker_direct_connection_deserialize_finish (TrackerSparqlConnection  *connection,
+                                              GAsyncResult             *res,
+                                              GError                  **error)
+{
+	return g_task_propagate_boolean (G_TASK (res), error);
+}
+
+static void
+tracker_direct_connection_map_connection (TrackerSparqlConnection *connection,
+					  const gchar             *handle_name,
+					  TrackerSparqlConnection *service_connection)
+{
+	TrackerDirectConnectionPrivate *priv;
+	TrackerDirectConnection *conn;
+
+	conn = TRACKER_DIRECT_CONNECTION (connection);
+	priv = tracker_direct_connection_get_instance_private (conn);
+
+	tracker_data_manager_map_connection (priv->data_manager,
+	                                     handle_name,
+	                                     service_connection);
+}
+
 static void
 tracker_direct_connection_class_init (TrackerDirectConnectionClass *klass)
 {
@@ -1210,6 +1511,12 @@ tracker_direct_connection_class_init (TrackerDirectConnectionClass *klass)
 	sparql_connection_class->update_resource_async = tracker_direct_connection_update_resource_async;
 	sparql_connection_class->update_resource_finish = tracker_direct_connection_update_resource_finish;
 	sparql_connection_class->create_batch = tracker_direct_connection_create_batch;
+	sparql_connection_class->lookup_dbus_service = tracker_direct_connection_lookup_dbus_service;
+	sparql_connection_class->serialize_async = tracker_direct_connection_serialize_async;
+	sparql_connection_class->serialize_finish = tracker_direct_connection_serialize_finish;
+	sparql_connection_class->deserialize_async = tracker_direct_connection_deserialize_async;
+	sparql_connection_class->deserialize_finish = tracker_direct_connection_deserialize_finish;
+	sparql_connection_class->map_connection = tracker_direct_connection_map_connection;
 
 	props[PROP_FLAGS] =
 		g_param_spec_flags ("flags",
@@ -1237,21 +1544,50 @@ tracker_direct_connection_class_init (TrackerDirectConnectionClass *klass)
 	g_object_class_install_properties (object_class, N_PROPS, props);
 }
 
-TrackerDirectConnection *
+TrackerSparqlConnection *
 tracker_direct_connection_new (TrackerSparqlConnectionFlags   flags,
                                GFile                         *store,
                                GFile                         *ontology,
                                GError                       **error)
 {
-	g_return_val_if_fail (!store || G_IS_FILE (store), NULL);
-	g_return_val_if_fail (!ontology || G_IS_FILE (ontology), NULL);
-	g_return_val_if_fail (!error || !*error, NULL);
+	return g_initable_new (TRACKER_TYPE_DIRECT_CONNECTION,
+	                       NULL, error,
+	                       "flags", flags,
+	                       "store-location", store,
+	                       "ontology-location", ontology,
+	                       NULL);
+}
 
-	return g_object_new (TRACKER_TYPE_DIRECT_CONNECTION,
-	                     "flags", flags,
-	                     "store-location", store,
-	                     "ontology-location", ontology,
-	                     NULL);
+void
+tracker_direct_connection_new_async (TrackerSparqlConnectionFlags  flags,
+                                     GFile                        *store,
+                                     GFile                        *ontology,
+                                     GCancellable                 *cancellable,
+                                     GAsyncReadyCallback           cb,
+                                     gpointer                      user_data)
+{
+	g_async_initable_new_async (TRACKER_TYPE_DIRECT_CONNECTION,
+	                            G_PRIORITY_DEFAULT,
+	                            cancellable,
+	                            cb,
+	                            user_data,
+	                            "flags", flags,
+	                            "store-location", store,
+	                            "ontology-location", ontology,
+	                            NULL);
+}
+
+TrackerSparqlConnection *
+tracker_direct_connection_new_finish (GAsyncResult  *res,
+                                      GError       **error)
+{
+	GAsyncInitable *initable;
+
+	initable = g_task_get_source_object (G_TASK (res));
+
+	return TRACKER_SPARQL_CONNECTION (g_async_initable_new_finish (initable,
+	                                                               res,
+	                                                               error));
 }
 
 TrackerDataManager *
